<?php
/**
mnemo2svx.php by Rodrigo Severo

mnemo2svx.php converts Excel files produced by MNemoBridge
(http://arianesline.azurewebsites.net/DownloadAriane/IndexLocal) to Survex files
(https://survex.com/) that can also be used with Therion (https://therion.speleo.sk/).

Copyright (C) 2021  Rodrigo Severo <rsev@pm.me>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

define('MNEMO2SVX_VERSION', 0.3);
define('MAX_DELTA_AZIMUT', 5);
define('MAX_DELTA_DEPTH', 0.2);
define('AZIMUTH_FOR_VERTICAL_INTERMEDIATE_LEGS', 0);
define('DEFAULT_FIRST_STATION_NUMBER', 1);

require 'vendor/autoload.php';
use PhpOffice\PhpSpreadsheet\Spreadsheet;

require_once('2svx.inc.php');

function help() {
  global $argv;

  printf(<<<HELP
Usage:
  %s XLSX_MNEMO_BRIDGE_GENERATED_FILE_INPUT SURVEY_NAME [FIRST_STATION_NUMBER]

Version:
  %s


mnemo2svx.php  Copyright (C) 2021  Rodrigo Severo <rsev@pm.me>
This program comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under certain conditions.

To get a copy of the GNU General Public License see <https://www.gnu.org/licenses/>.


HELP
    ,
    $argv[0],
    MNEMO2SVX_VERSION
  );
}

function azimuth_mean($angle1, $angle2) {
  $y_part = $x_part = 0;
  $angles = [$angle1, $angle2];
	$size = count($angles);
	for ($i = 0; $i < $size; $i++) {
		$x_part += cos(deg2rad($angles[$i]));
		$y_part += sin(deg2rad($angles[$i]));
	}
	$x_part /= $size;
	$y_part /= $size;

  $mean = rad2deg(atan2($y_part, $x_part));

  /**
   * Transforming negative azimuths to positive ones.
   */
  if ($mean < 0) {
    $mean += 360;
  }

	return number_format($mean, 1, '.', '');
}

function azimuth_delta($angle1, $angle2) {
  $angles = [$angle1, $angle2];
  rsort($angles, SORT_NUMERIC);
  $delta = $angles[0] - $angles[1];
  if ($delta > 180) {
    $angles[0] = $angles[0] - 360;
    $delta = $angles[1] - $angles[0];
  }

  return $delta;
}

function depth_delta($DepthIn, $DepthOut) {
  $depths = [$DepthIn, $DepthOut];
  rsort($depths, SORT_NUMERIC);
  $delta = $depths[0] - $depths[1];

  return $delta;
}

function depth_comment($delta, $DepthIn, $DepthOut) {
  $comments = [];
  if ($delta > MAX_DELTA_DEPTH) {
    $comments[] = sprintf(
      'Depth delta WARNING: delta greater than %01.2f (%01.2f) between previous DepthOut (%01.2f) and current DepthIn (%01.2f)',
      MAX_DELTA_DEPTH,
      $delta,
      $DepthOut,
      $DepthIn
    );
    $comments[] = '-> Intermediate leg created below.';
  }
  else {
    // $comments[] = sprintf(
    //   'Depth delta (%01.2f) between previous DepthOut (%01.2f) and current DepthIn (%01.2f)',
    //   $delta,
    //   $DepthIn,
    //   $DepthOut
    // );
  }
  return $comments;
}

function azimuth_comment($AzimuthMean, $AzimuthDelta, $AzimuthIn, $AzimuthOut) {
  $comments = [];
  if ($AzimuthDelta > MAX_DELTA_AZIMUT) {
    $comments[] = sprintf(
      'Azimuth WARNING: difference between IN (%01.1f) and OUT (%01.1f) azimuths greater than limit (%01.1f): %01.1f',
      $AzimuthIn,
      $AzimuthOut,
      MAX_DELTA_AZIMUT,
      $AzimuthDelta
    );
  }
  // $comments[] = sprintf('Adopted azimuth mean: %s', $AzimuthMean);

  return $comments;
}

function azimuth_normal($AzFrom, $AzTo) {
  // $t = array(
  //   'AzFrom' => $AzFrom,
  //   'AzTo' => $AzTo
  // );

  // Calculating the difference of both azimuths.
  $delta = $AzTo - $AzFrom;

  // Making the difference be across the sortest path.
  if ($delta > 180) {
    $delta -= 360;
  }
  elseif ($delta < -180) {
    $delta += 360;
  }

  // Calculating the mean azimuth.
  $AzMean = $AzFrom + ($delta / 2);

  // Calculating the normal of the mean azimuth.
  if ($delta > 0) {
    $AzNormal = $AzMean + 90;
  }
  else {
    $AzNormal = $AzMean - 90;
  }

  // Reducing the normal azimith to the 0°/360° range.
  if ($AzNormal > 360) {
    $AzNormal -= 360;
  }
  elseif ($AzNormal < 0) {
    $AzNormal += 360;
  }

  // $t['delta'] = $delta;
  // $t['AzPer'] = $AzPer;
  // print_r($t);

  return $AzNormal;
}


function read_xlsx_file($d, $firstStationCreated) {
  $data = array();

  $data['infile_name'] = trim($d[1][0]);

  $unit = $d[2][1];
  $data['unit'] = trim(substr(trim($unit), 6));

  $date = $d[3][0];
  $data['date'] = trim(substr(trim($date), 6));

  $measurements = array_slice($d, 6);
  // print_r($measurements);
  /**
   * Starting ID most be one higher than the firstStationCreated as the
   * firstStationCreated will be the fromID.
   */
  $ID = $firstStationCreated + 1;
  $firstMeasurement = true;
  $intermediateStation = null;
  foreach ($measurements as $line) {
    if (strtolower($line[0]) != 'std') {
      continue;
    }

    $AzimuthIn = trim($line[2]);
    $AzimuthOut = trim($line[3]);
    $AzimuthMean = azimuth_mean($AzimuthIn, $AzimuthOut);
    $AzimuthDelta = azimuth_delta($AzimuthIn, $AzimuthOut);
    $AzimuthComments = azimuth_comment($AzimuthMean, $AzimuthDelta, $AzimuthIn, $AzimuthOut);
    $PitchIn = trim($line[4]);
    $PitchOut = trim($line[5]);
    $DepthIn = trim($line[6]);
    $DepthOut = trim($line[7]);
    if ($firstMeasurement) {
      $firstMeasurement = false;
      $DepthDelta = 0;
      $DepthComments = [];
      $DepthFrom = $DepthIn;
      $AzimuthNormalMeanPrevious = 0;
    }
    else {
      $DepthDelta = depth_delta(
        $DepthIn,
        $data['measurements'][$ID - 1]['DepthTo']
      );
      $DepthComments = depth_comment(
        $DepthDelta,
        $DepthIn,
        $data['measurements'][$ID - 1]['DepthTo']
      );
      if ($DepthDelta > MAX_DELTA_DEPTH) {
        $DepthFrom = $DepthIn;
      }
      else {
        $DepthFrom = ($DepthIn + $data['measurements'][$ID - 1]['DepthOut']) / 2;
        $data['measurements'][$ID - 1]['DepthTo'] = $DepthFrom;
      }
      $AzimuthNormalMeanPrevious = azimuth_normal(
        $data['measurements'][$ID - 1]['Azimuth'],
        $AzimuthMean,
      );
    }

    if ($DepthDelta > MAX_DELTA_DEPTH) {
      $intermediateStation = sprintf('%uB', $ID - 1);
      $data['intermediates'][$ID] = array(
        'fromID' => $ID - 1,
        'ID' => $intermediateStation,
        'Length' => $DepthDelta,
        'DepthPrevious' => $data['measurements'][$ID - 1]['DepthTo'],
        'DepthFinal' => $DepthFrom,
      );
    }
    else {
      $intermediateStation = null;
    }

    $data['measurements'][$ID] = array(
      'fromID' => is_null($intermediateStation) ? ($ID - 1) : $intermediateStation,
      'ID' => $ID,
      'Length' => trim($line[1]),
      'AzimuthIn' => $AzimuthIn,
      'AzimuthOut' => $AzimuthOut,
      'PitchIn' => $PitchIn,
      'PitchOut' => $PitchOut,
      'DepthIn' => $DepthIn,
      'DepthOut' => $DepthOut,
      'DepthFrom' => $DepthFrom,
      'DepthTo' => $DepthOut,
      'Azimuth' => $AzimuthMean,
      'AzimuthDelta' => $AzimuthDelta,
      'AzimuthNormalMeanPrevious' => $AzimuthNormalMeanPrevious,
      'AzimuthComments' => $AzimuthComments,
      'DepthComments' => $DepthComments,
    );

    $lastStationCreated = $ID;
    $ID++;
  }
  $data['export'] = sprintf(
    '%s %s',
    $firstStationCreated,
    trim($lastStationCreated)
  );

  return $data;
}
function header_comments() {
  global $xlsx_filename;

  $header = sprintf(<<<CONTENTS
; svx initially file created:
; * with mnemo2svx.php version %s
; * from "%s"
; * at %s


CONTENTS
    ,
    MNEMO2SVX_VERSION,
    $xlsx_filename,
    date('c')
  );

  return $header;
}

function get_svx_contents($data, $survey_name) {
  $survey_name = trim($survey_name);
  $survey_slug = slugify($survey_name);

  $contents = header_comments();

  $prefix = '';
  $contents .= new_line(sprintf('*begin %s', $survey_slug), $prefix);

  $prefix = "\t";
  $contents .= new_line(
    '; First and last stations automatically exported',
    $prefix
  );
  $contents .= new_line(sprintf('*export %s', trim($data['export'])), $prefix);
  $contents .= "\n";

  $contents .= new_line(sprintf('*title "%s"', $survey_name), $prefix);
  $contents .= new_line(sprintf(
    '; original survey name: "%s"', $data['infile_name']),
    $prefix
  );
  $contents .= "\n";

  if (isset($data['date'])) {
    $contents .= new_line(
      sprintf('*date %s', date_in_svx_format($data['date'])),
      $prefix
    );
    $contents .= "\n";
  }

  $contents .= new_line('; Uncomment and fill the lines below to set the team members names.', $prefix);
  $contents .= new_line(';*team "" explorer', $prefix);
  $contents .= new_line(';*team "" surveyor', $prefix);
  $contents .= "\n";

  $contents .= new_line('*require 1.2.21', $prefix);
  $contents .= "\n";

  $contents .= new_line('*case preserve', $prefix);
  $contents .= "\n";

  $contents .= new_line('*instrument compass MNemo', $prefix);
  $contents .= new_line('*instrument clino MNemo', $prefix);
  $contents .= new_line('*instrument tape MNemo', $prefix);
  $contents .= "\n";

  $contents .= new_line('*calibrate depth 0 -1', $prefix);
  $contents .= "\n";

  if (isset($data['unit'])) {
    $unit = strtolower($data['unit']);
    if ($unit == 'metric') {
      $contents .= new_line('*units tape depth metres', $prefix);
    }
    else {
      $contents .= new_line('*units tape depth feet', $prefix);
    }
    $contents .= "\n";
  }

  $contents .= new_line('; --------------------------------------------------', $prefix);
  $contents .= new_line('; Main topo data', $prefix);
  $contents .= new_line('; --------------------------------------------------', $prefix);
  $contents .= new_line(
    '*data diving from to tape compass fromdepth todepth ignoreall',
    $prefix
  );
  $contents .= new_line(
    "; From\tTo\tLength\tAzimuth\tFromDep\tToDep\tAzIn\tAzOut\tAzDelta\tAzNormal\tDepIn\tDepOut\tPitchIn\tPitchOut",
    $prefix
  );
  $contents .= "\n";
  $firstLine = true;
  foreach ($data['measurements'] as $m) {
    if ((count($m['AzimuthComments']) > 0) || (count($m['DepthComments']) > 0)) {
      if (!$firstLine) {
        $contents .= "\n";
      }
    }
    if (count($m['AzimuthComments']) > 0) {
      foreach ($m['AzimuthComments'] as $comment) {
        $contents .= new_line(sprintf('; %s', $comment), $prefix);
      }
    }
    if (count($m['DepthComments']) > 0) {
      foreach ($m['DepthComments'] as $comment) {
        $contents .= new_line(sprintf('; %s', $comment), $prefix);
      }
    }
    $contents .= new_line(
      sprintf(
        "%s\t%s\t%01.2f\t%01.1f\t%01.2f\t%01.2f\t%01.1f\t%01.1f\t%01.1f\t%01.1f\t%01.2f\t%01.2f\t%01.1f\t%01.1f",
        trim($m['fromID']),
        trim($m['ID']),
        $m['Length'],
        $m['Azimuth'],
        $m['DepthFrom'],
        $m['DepthTo'],
        $m['AzimuthIn'],
        $m['AzimuthOut'],
        $m['AzimuthDelta'],
        $m['AzimuthNormalMeanPrevious'],
        $m['DepthIn'],
        $m['DepthOut'],
        $m['PitchIn'],
        $m['PitchOut']
      ),
      $prefix
    );
    $firstLine = false;
  }
  $contents .= "\n";

  if (isset($data['intermediates']) &&
    is_array($data['intermediates']) &&
    (count($data['intermediates']) > 0)) {

    $contents .= new_line('; --------------------------------------------------', $prefix);
    $contents .= new_line('; Intermediate vertical legs', $prefix);
    $contents .= new_line('; --------------------------------------------------', $prefix);
    $contents .= new_line(
      '*data normal from to tape compass clino ignoreall',
      $prefix
    );
    $contents .= new_line(
      "; From\tTo\tLength\tAzimuth\tUp/Down\tFromDep\tToDep",
      $prefix
    );
    $contents .= "\n";

    foreach ($data['intermediates'] as $i) {
      $direction = ($i['DepthPrevious'] < $i['DepthFinal']) ? 'DOWN' : 'UP';
      $contents .= new_line(
        sprintf(
          "%s\t%s\t%01.2f\t-\t%s\t%01.2f\t%01.2f",
          trim($i['fromID']),
          trim($i['ID']),
          $i['Length'],
          $direction,
          $i['DepthPrevious'],
          $i['DepthFinal']
        ),
        $prefix
      );
    }
    $contents .= "\n";
  }

  $prefix = '';
  $contents .= new_line(sprintf('*end %s', $survey_slug), $prefix);

  return $contents;
}

function get_output_filename($survey_name, $input) {
  $parts = pathinfo($input);
  $slug_name = slugify($survey_name);
  $output = sprintf('%s/%s.svx', $parts['dirname'], $slug_name);

  return $output;
}

if (($argc < 3) || ($argc > 4)) {
  help();
  die();
}

$xlsx_filename = $argv[1];
$survey_name = $argv[2];
$first_station = ($argc == 4) ? (int) $argv[3] : DEFAULT_FIRST_STATION_NUMBER;

$reader = new \PhpOffice\PhpSpreadsheet\Reader\Xlsx();

$spreadsheet = $reader->load($xlsx_filename);
$d = $spreadsheet->getSheet(0)->toArray();
unset($spreadsheet);
unset($reader);

// print_r($d);

// echo 'Média de 350 e 10: ' . azimut_delta(350, 10) . "\n";
// echo 'Média de 10 e 350: ' . azimut_delta(10, 350) . "\n";
// echo 'Média de 20 e 10: ' . azimut_delta(20, 10) . "\n";
// echo 'Média de 10 e 20: ' . azimut_delta(10, 20) . "\n";
// echo 'Média de 176 e 181: ' . azimut_delta(176, 181) . "\n";
// echo 'Média de 17 e 279: ' . azimut_delta(17, 279) . "\n";
//
// die();

$data = read_xlsx_file($d, $first_station);
// print_r($data);
$contents = get_svx_contents($data, $survey_name);
// print_r($contents);
$output_filename = get_output_filename($survey_name, $xlsx_filename);
// print_r($output_filename);
file_put_contents($output_filename, $contents);
