<?php
/**
tml2svx.php by Rodrigo Severo

tml2svx.php converts TMl files produced by Ariane
(http://arianesline.azurewebsites.net/Home/Ariane) to Survex files
(https://survex.com/) that can also be used with Therion (https://therion.speleo.sk/).

Copyright (C) 2021  Rodrigo Severo <rsev@pm.me>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

define('TML2SVX_VERSION', 0.4);
define('DEFAULT_ALTITUDE', 0);

require_once('coordinates.php');
require_once('2svx.inc.php');

function help() {
  global $argv;

  printf(<<<HELP
Usage:
  %s TML_FILE

Version:
  %s


tml2svx.php  Copyright (C) 2021  Rodrigo Severo <rsev@pm.me>
This program comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under certain conditions.

To get a copy of the GNU General Public License see <https://www.gnu.org/licenses/>.


HELP
    ,
    $argv[0],
    TML2SVX_VERSION
  );
}

function parse_array_or_string($input) {
  if (is_array($input)) {
    $output = '';
    foreach ($input as $input_line) {
      $output .= "/n" . trim($input_line);
    }
    if ($output != '') {
      $output = substr($output, 2);
    }
  }
  else {
    $output = trim($input);
  }

  return $output;
}

function parse_tml_file($tml) {
  $data = array();
  if (isset($tml['caveName'])) {
    $caveName = trim($tml['caveName']);
    if ($caveName != '') {
      $data['caveName'] = $caveName;
    }
  }
  if (isset($tml['firstStartAbsoluteElevation']) && ($tml['firstStartAbsoluteElevation'] != 0)) {
    $data['firstStartAbsoluteElevation'] = $tml['firstStartAbsoluteElevation'];
  }
  if (isset($tml['geoCoding'])) {
    $geocoding = trim($tml['geoCoding']);
    if ($geocoding != '') {
      $data['geoCoding'] = $geocoding;
    }
  }
  if (isset($tml['unit'])) {
    $unit = trim($tml['unit']);
    if ($unit != '') {
      $data['unit'] = $unit;
    }
  }
  foreach ($tml['Data']['SurveyData'] as $surveydata) {
    if (strtolower($surveydata['Type']) == 'closure') {
      continue;
    }
    $block = trim($surveydata['Section']);
    if (($surveydata['FromID'] != -1) && ($surveydata['ID'] != -1)) {
      $fromID = trim($surveydata['FromID']);
      $ID = trim($surveydata['ID']);
      $stations = sprintf('%s|%s', $fromID, $ID);
      $data['blocks_of_station'][$fromID][$block] = true;
      $data['blocks_of_station'][$ID][$block] = true;
      $data['blocks'][$block]['measurements'][$stations]['FromID'] = $fromID;
      $data['blocks'][$block]['measurements'][$stations]['ID'] = $ID;
      $data['blocks'][$block]['measurements'][$stations]['Azimut'] = $surveydata['Azimut'];
      if (isset($surveydata['Color'])) {
        $color = trim($surveydata['Color']);
        if ($color != '') {
          $data['blocks'][$block]['measurements'][$stations]['Color'] = $color;
        }
      }
      if (isset($surveydata['Comment'])) {
        $comment = parse_array_or_string($surveydata['Comment']);
        if ($comment != '') {
          $data['blocks'][$block]['measurements'][$stations]['Comment'] = $comment;
        }
      }
      if (isset($surveydata['Date'])) {
        $date = trim($surveydata['Date']);
        if ($date != '') {
          $data['blocks'][$block]['Dates'][$date] = true;
          $data['Dates'][$date] = true;
        }
      }
      $data['blocks'][$block]['measurements'][$stations]['Depth'] = $surveydata['Depth'];
      // Setting all known depths.
      get_set_station_depth($ID, trim($surveydata['Depth']));
      $data['blocks'][$block]['measurements'][$stations]['Down'] = $surveydata['Down'];
      $data['blocks'][$block]['measurements'][$stations]['Excluded'] = ($surveydata['Excluded'] != 'false');
      if (isset($surveydata['Explorer'])) {
        if (is_array($surveydata['Explorer'])) {
          if (isset($surveydata['Explorer']['Explorer'])) {
            $explorer = trim($surveydata['Explorer']['Explorer']);
            if (($explorer != '?') && ($explorer != '')) {
              $data['blocks'][$block]['team']['Explorer'] = $explorer;
            }
          }
          if (isset($surveydata['Explorer']['Surveyor'])) {
            $surveyor = trim($surveydata['Explorer']['Surveyor']);
            if (($surveyor != '?') && ($surveyor != '')) {
              $data['blocks'][$block]['team']['Surveyor'] = $surveyor;
            }
          }
        }
        else {
          $explorer = preg_match(
            '_<Explorer>(.*)</Explorer>_',
            $surveydata['Explorer'],
            $matchs
          );
          if ($explorer) {
            $explorer = trim($matchs[1]);
            $data['blocks'][$block]['team']['Explorer'] = $explorer;
          }
          $surveyor = preg_match(
            '_<Surveyor>(.*)</Surveyor>_',
            $surveydata['Explorer'],
            $matchs
          );
          if ($surveyor) {
            $surveyor = trim($matchs[1]);
            $data['blocks'][$block]['team']['Surveyor'] = $surveyor;
          }
        }
      }
      $data['blocks'][$block]['measurements'][$stations]['Left'] = $surveydata['Left'];
      $data['blocks'][$block]['measurements'][$stations]['Length'] = $surveydata['Length'];
      if (isset($surveydata['Name'])) {
        $name = parse_array_or_string($surveydata['Name']);
        if ($name != '') {
          $data['blocks'][$block]['measurements'][$stations]['Name'] = $name;
        }
      }
      $data['blocks'][$block]['measurements'][$stations]['Right'] = $surveydata['Right'];
      $data['blocks'][$block]['measurements'][$stations]['Up'] = $surveydata['Up'];
    }
    if ($surveydata['ID'] != -1) {
      if (($surveydata['Latitude'] != 0) || ($surveydata['Longitude'] != 0)) {
        $data['blocks'][$block]['fixes'][] = array(
          'ID' => trim($surveydata['ID']),
          'Latitude' => $surveydata['Latitude'],
          'Longitude' => $surveydata['Longitude'],
        );
        if (!isset($data['declinationFix'])) {
          $data['declinationFix'] = array(
            'Latitude' => $surveydata['Latitude'],
            'Longitude' => $surveydata['Longitude'],
          );
        }
      }
      if (strtolower($surveydata['Type']) == 'start') {
        $data['blocks'][$block]['entrances'][] = $surveydata['ID'];
      }
    }
  }

  return $data;
}

function identify_equates_and_exports($data) {
  foreach ($data['blocks_of_station'] as $id => $blocks) {
    if (count($blocks) == 1) {
      continue;
    }
    $data['equates'][$id] = '';
    foreach ($blocks as $block => $temp) {
      $data['equates'][$id] .= sprintf('%s.%u ', slugify($block), $id);
      if (!isset($data['blocks'][$block]['export'])) {
        $data['blocks'][$block]['export'] = '';
      }
      $data['blocks'][$block]['export'] .= sprintf('%u ', $id);
    }
  }

  return $data;
}

function svx_filename_from_tml_filename($tml_filename) {
  $parts = explode('.', $tml_filename);
  if (count($parts) > 1) {
    array_pop($parts);
  }
  $svx_filename = implode('.', $parts) . '.svx';

  return $svx_filename;
}

function get_set_station_depth($id, $depth = null) {
  static $depths = array();

  if (is_null($depth)) {
    if (isset($depths[$id])) {
      return $depths[$id];
    }
    else {
      return '0';
    }
  }
  else {
    $depths[$id] = $depth;
    return $depth;
  }
}

function header_comments() {
  global $tml_filename;

  $header = sprintf(<<<CONTENTS
; svx initially file created:
; * with tml2svx.php version %s
; * from "%s"
; * at %s


CONTENTS
    ,
    TML2SVX_VERSION,
    $tml_filename,
    date('c')
  );

  return $header;
}

function write_svx_file($file, $data) {
  $contents = header_comments();

  $nome_caverna = slugify($data['caveName']);
  $prefix = '';
  $contents .= new_line(sprintf('*begin %s', $nome_caverna), $prefix);
  $prefix = "\t";
  $contents .= new_line(sprintf('*title "%s"', $data['caveName']), $prefix);
  if (isset($data['geoCoding'])) {
    $contents .= new_line(sprintf('; geoCoding "%s"', $data['geoCoding']), $prefix);
  }
  $contents .= "\n";

  $contents .= new_line('*require 1.2.21', $prefix);
  $contents .= "\n";

  $contents .= new_line('*case preserve', $prefix);
  $contents .= "\n";

  $contents .= new_line('*instrument compass MNemo', $prefix);
  $contents .= new_line('*instrument clino MNemo', $prefix);
  $contents .= new_line('*instrument tape MNemo', $prefix);
  $contents .= "\n";

  if (isset($data['unit'])) {
    $unit = strtolower($data['unit']);
    if ($unit == 'ft') {
      $contents .= new_line('*units tape depth feet', $prefix);
    }
    else {
      $contents .= new_line('*units tape depth metres', $prefix);
    }
    $contents .= "\n";
  }

  $contents .= new_line('*calibrate depth 0 -1', $prefix);
  if (isset($data['firstStartAbsoluteElevation'])) {
    $contents .= new_line(
      sprintf('*calibrate depth ', $data['firstStartAbsoluteElevation']),
      $prefix
    );
  }
  $contents .= "\n";

  if (isset($data['declinationFix'])) {
    $lat = trim($data['declinationFix']['Latitude']);
    $long = trim($data['declinationFix']['Longitude']);

    $contents .= new_line('*cs LONG-LAT', $prefix);
    $utm = ll2utm($lat, $long);
    if ($utm['success']) {
      $utm = $utm['attr'];
      $hemisphere = $utm['aboveEquator'] ? 'N' : 'S';
      $contents .= new_line(
        sprintf('*cs out UTM%u%s', $utm['zone'], $hemisphere),
        $prefix
      );
    }
    else {
      $contents .= new_line('*cs out S-MERC', $prefix);
    }
    $contents .= "\n";

    $contents .= new_line(
      sprintf('*declination auto %s %s %s', $long, $lat, DEFAULT_ALTITUDE),
      $prefix
    );
    $contents .= "\n";
  }

  if (isset($data['Dates'])) {
    $dates = array_keys($data['Dates']);
    if (count($dates) == 1) {
      $contents .= new_line(
        sprintf('*date %s', date_in_svx_format($dates[0])),
        $prefix
      );
    }
    else {
      sort($dates);
      $last_date = array_pop($dates);
      $contents .= new_line(
        sprintf(
          '*date %s-%s',
          date_in_svx_format($dates[0]),
          date_in_svx_format($last_date)
        ),
        $prefix
      );
    }
    $contents .= "\n";
  }

  if (isset($data['equates'])) {
    foreach ($data['equates'] as $equate) {
      $contents .= new_line(sprintf('*equate %s', trim($equate)), $prefix);
    }
    $contents .= "\n";
  }

  foreach ($data['blocks'] as $name => $block) {
    $slug = slugify($name);
    $contents .= new_line(sprintf('*begin %s', $slug), $prefix);
    $prefix = "\t\t";
    if (isset($block['export'])) {
      $contents .= new_line(
        sprintf('*export %s', trim($block['export'])),
        $prefix
      );
      $contents .= "\n";
    }

    if (isset($block['fixes'])) {
      foreach ($block['fixes'] as $fix) {
        $lat = trim($fix['Latitude']);
        $long = trim($fix['Longitude']);
        $contents .= new_line(
          sprintf(
            '*fix %s %s %s %s',
            trim($fix['ID']),
            $long,
            $lat,
            DEFAULT_ALTITUDE
          ),
          $prefix
        );
      }
      $contents .= "\n";
    }

    if (isset($block['entrances'])) {
      foreach ($block['entrances'] as $entrance) {
        $contents .= new_line(sprintf('*entrance %s', trim($entrance)), $prefix);
      }
      $contents .= "\n";
    }

    if (isset($block['Dates'])) {
      $dates = array_keys($block['Dates']);
      if (count($dates) == 1) {
        $contents .= new_line(
          sprintf('*date %s', date_in_svx_format($dates[0])),
          $prefix
        );
      }
      else {
        sort($dates);
        $last_date = array_pop($dates);
        $contents .= new_line(
          sprintf(
            '*date %s-%s',
            date_in_svx_format($dates[0]),
            date_in_svx_format($last_date)
          ),
          $prefix
        );
      }
      $contents .= "\n";
    }

    if (isset($block['team'])) {
      if ((isset($block['team']['Explorer']) &&
          isset($block['team']['Surveyor']) &&
          (strtolower(trim($block['team']['Explorer'])) == strtolower(trim($block['team']['Surveyor']))))
        )
      {
        $contents .= new_line(
          sprintf(
            '*team "%s" explorer surveyor',
            trim($block['team']['Explorer'])
          ),
          $prefix
        );
      }
      else {
        if (isset($block['team']['Explorer'])) {
          $contents .= new_line(
            sprintf(
              '*team "%s" explorer',
              trim($block['team']['Explorer'])
            ),
            $prefix
          );
        }
        if (isset($block['team']['Surveyor'])) {
          $contents .= new_line(
            sprintf(
              '*team "%s" surveyor',
              trim($block['team']['Surveyor'])
            ),
            $prefix
          );
        }
      }
    }
    else {
      $contents .= new_line(
        '; Uncomment and fill the lines below to set the team members responsible for this survey.',
        $prefix
      );
      $contents .= new_line(';*team "" explorer', $prefix);
      $contents .= new_line(';*team "" surveyor', $prefix);
    }
    $contents .= "\n";

    if (isset($block['measurements'])) {
      $contents .= new_line(
        '*data diving from to fromdepth todepth tape compass ignoreall',
        $prefix
      );
      $contents .= new_line(
        "; From\tTo\tFromDepth\tToDepth\tLength\tAzimuth\tColor\tName\tComments",
        $prefix
      );
      $contents .= "\n";

      foreach ($block['measurements'] as $measurement) {
        $name = '';
        if (isset($measurement['Name'])) {
          $name = trim($measurement['Name']);
        }
        if ($name == '') {
          $name = '-';
        }

        $comment = '';
        if (isset($measurement['Comment'])) {
          $comment = trim($measurement['Comment']);
        }
        if ($comment == '') {
          $comment = '-';
        }

        $color = '';
        if (isset($measurement['Color'])) {
          $color = trim($measurement['Color']);
        }
        if ($color == '') {
          $color = '-';
        }

        $fromID = trim($measurement['FromID']);
        $id = trim($measurement['ID']);
        $contents .= new_line(
          sprintf(
            "%s\t%s\t%01.2f\t%01.2f\t%01.2f\t%01.1f\t%s\t%s\t%s",
            $fromID,
            $id,
            get_set_station_depth($fromID),
            get_set_station_depth($id),
            trim($measurement['Length']),
            trim($measurement['Azimut']),
            $color,
            $name,
            $comment,
          ),
          $prefix
        );
      }
      $contents .= "\n";

      $contents .= new_line(
        '*data passage station left right up down',
        $prefix
      );
      $contents .= new_line(
        "; Station\tLeft\tRight\tUp\tDown",
        $prefix
      );
      $contents .= "\n";

      foreach ($block['measurements'] as $measurement) {
        $contents .= new_line(
          sprintf(
            "%s\t%01.2f\t%01.2f\t%01.2f\t%01.2f",
            trim($measurement['ID']),
            trim($measurement['Left']),
            trim($measurement['Right']),
            trim($measurement['Up']),
            trim($measurement['Down']),
          ),
          $prefix
        );
      }
      $contents .= "\n";
    }
    $prefix = "\t";
    $contents .= new_line(sprintf('*end %s', $slug), $prefix);
    $contents .= "\n";
  }

  $prefix = '';
  $contents .= new_line(sprintf('*end %s', $nome_caverna), $prefix);

  file_put_contents($file, $contents);
}

if ($argc != 2) {
  help();
  die();
}
$tml_filename = $argv[1];
$tml_file = shell_exec("zcat $tml_filename");
$tml_file = simplexml_load_string($tml_file, 'SimpleXMLElement', LIBXML_NOCDATA);
$tml_file = json_encode($tml_file);
$tml_file = json_decode($tml_file, TRUE);
// print_r($tml_file);
$svx_data = parse_tml_file($tml_file);
$svx_data = identify_equates_and_exports($svx_data);
// print_r($svx_data);
$svx_filename = svx_filename_from_tml_filename($tml_filename);
write_svx_file($svx_filename, $svx_data);
